/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.runtime;

import jasco.runtime.aspect.IHook;
import jasco.runtime.hotswap.JAsCoClassLoader;
import jasco.runtime.transform.TransformerConstants;
import jasco.util.logging.Logger;

import java.io.Serializable;
import java.lang.annotation.Annotation;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Vector;

import javassist.CtClass;
import javassist.CtMethod;
import javassist.NotFoundException;


/**
 * The calledmethod keyword availabe in JAsCo aspects is of type JascoMethod, this means that all subsequent methods are availabe for this keyword.
 * calledmethod refers to the method joinpoint that triggered the aspect.
 */
//LDBN 6-6-2005 Make the class Serializable
public class JascoMethod implements Cloneable, MethodJoinpoint, Serializable {
	
	private static final IHook[] EMPTY_LINKED_HOOKS = new IHook[0];
	
	private String name = "";
	private String className = "className";
	private Object[] args = null;
	private Object calledobject = null;
	private int methodid = -1;
	private transient JAsCoBeanCallback callback = null;
	//private JascoMethod[] copies;
	private transient CombinedHooks hooksInvoker = null;
	private int modifiers = 0;
	private transient IHook[] linkedHooks = EMPTY_LINKED_HOOKS;
	private transient int currentHook = 0;
	private String fullname = "";
	private transient boolean inlineCompiled = false;
	private boolean isExecution = true;
	private String[] fargs = null;
	private String[] excs = null;
	private int line = -1;
	
	private transient ClassLoader loader = ClassLoader.getSystemClassLoader();
	private transient java.lang.reflect.Method thisMethod =null;
	private String returnTypeName;
	
	private transient boolean dontLoadClass=false;
	private transient CtMethod javassistMethod;
	/**
	*  Do not use, internal usage only
	*/ 
	public JascoMethod()  {
	}
	
	/**
	*  Do not use, internal usage only
	*/
	public JascoMethod(String aName)  
	{
		name = aName;	
	}
	
	
	/**
	*  Do not use, internal usage only
	*/
	public JascoMethod(String fullname, String aName, String aClassName, int methodid, JAsCoBeanCallback callbackObject, int modifiers)  
	{
		this(fullname,aName,aClassName,methodid,callbackObject,modifiers,true);
	}
	
	/**
	*  Do not use, internal usage only
	* for hotswap 1, deprecated...
	*/
	public JascoMethod(String fullname,  Object object, Object[] args, String aName, String aClassName, int methodid, JAsCoBeanCallback callbackObject, int modifiers)  
	{
		this(fullname,aName,aClassName,methodid,callbackObject,modifiers,true);
		if(object!=null&&!object.equals(""))
			this.setCalledObject(object);
		this.setArgumentsArray(args);
	}
	
	/**
	*  Do not use, internal usage only
	*/
	public JascoMethod(String fullname,  String aName, String aClassName, int methodid, JAsCoBeanCallback callbackObject, int modifiers, boolean isExecution)  
	{
		name = aName;
		className = aClassName;
		this.methodid=methodid;
		callback = callbackObject;
		this.modifiers=modifiers;
		this.fullname=fullname;
		this.isExecution=isExecution;
	}
	
	public void setFormalArgumentTypes(String[] fargs) {
		
		this.fargs=fargs;
	}
	
	public void setExceptionTypes(String[] excs) {
		
		this.excs=excs;
	}
	
	protected Method getReflectMethod() {
		
		if(thisMethod==null)
			initReflectMethod();
		return thisMethod;
	}
	
	/*
	 *  (non-Javadoc)
	 * @see jasco.runtime.MethodJoinpoint#loadClass()
	 */
	public Class loadClass() {
		
		try {
		return getClassLoader().loadClass(this.getClassName());
		}catch(Exception ex) {
 			Logger.getInstance().showError(ex);
		}
		return null;
	}
	
	/**
	 *  Returns all defined annotations
	 */
	public Annotation[] getAnnotations() {
		
		return getReflectMethod().getAnnotations();
	}
	
	public Class getReturnType() {
		
		if(getReturnTypeName()==null)
			return getReflectMethod().getReturnType();
		
		try {
			return loadClass(getReturnTypeName());
		} catch (ClassNotFoundException e) {
			Logger.getInstance().showError(e);
		}
		return null;
	}
	
	
	protected void initReflectMethod() {
		
		initReflectMethod(loadClass());
		
	}
	
	
	protected void initReflectMethod(Class clz) {
		try {
			if(fargs==null) { //HotSwap 1
				Method[] methods = clz.getDeclaredMethods();
				for(int i=0;i<methods.length;++i)
					if(methods[i].getName().equals(getName())) {
						thisMethod=methods[i];
						break;
					}
				
			} else {
				thisMethod=clz.getDeclaredMethod(this.getName(), getFormalArgumentTypes());
			}

			thisMethod.setAccessible(true);
			
			}catch(Exception ex) {
	 			Logger.getInstance().showError(ex);
			}
			
	 }
	
	/**
	*  Returns the unique name for this method joinpoint
	*/
	public String getFullName()  {
	
		return fullname;
	}
	
	/**
	*  Returns the unique name for this method joinpoint
	*/
	public String getSignature()  {
	
		return fullname;
	}
	
	
	public void initClassLoader(ClassLoader loader) {
		
		this.loader=loader;
	}
	
	public String toString()  {
	
		return getFullName();
	}
	
	/**
	*  Returns true when this method joinpoint is an execution joinpoint
	*/
	public boolean isExecution()  {
	
		return isExecution;
	}
	
	/**
	*  Returns true when this method joinpoint is a call joinpoint
	*/
	public boolean  isCall()  {
	
		return !isExecution;
	}
	
	/**
	*  Returns the method name
	*/
	public String getName()  {
	
		return name;
	}
	
	public long getID()  {
	
		return methodid;
	}
	
	/**
	*  Returns the actual argument types
	*/
	public Class[] getActualArgumentTypes()  {
		
		Class[] formArgs = this.getFormalArgumentTypes();
		Class[] clz = new Class[getArgumentsArray().length];
		for(int i=0;i<clz.length;++i)
			if(formArgs[i].isPrimitive()) //formargs is primitive, actual args is boxed primite, but return primitive!
				clz[i]=formArgs[i];
			else clz[i]=getArgumentsArray()[i].getClass();
		
		return clz;
	}
	
	public ClassLoader getClassLoader() {
		
		if(loader==null)
			loader=ClassLoader.getSystemClassLoader();
		return loader;
	}
	
	protected Class loadClass(String aClass) throws ClassNotFoundException {
		
		int index = aClass.indexOf("[");
		if(index!=-1) {//arrays
			String name = aClass.substring(0,index);
			int num = (aClass.length()-index)/2;
			String res="";
			for(int i=0;i<num;++i)
				res+="[";
			res+=jasco.tools.jascoparser.JascoParser.getClassCode(name);
			return getClassLoader().loadClass(res);
		}
			
		if(aClass.equals("boolean"))
			return Boolean.TYPE;
		else if(aClass.equals("int"))
			return Integer.TYPE;
		else if(aClass.equals("float"))
			return Float.TYPE;
		else if(aClass.equals("double"))
			return Double.TYPE;
		else if(aClass.equals("char"))
			return Character.TYPE;
		else if(aClass.equals("byte"))
			return Byte.TYPE;
		else if(aClass.equals("short"))
			return Short.TYPE;
		else if(aClass.equals("long"))
			return Long.TYPE;
		else if(aClass.equals("void"))
			return Void.TYPE;	
		else return getClassLoader().loadClass(aClass);
	}
	
	/**
	*  Returns the formal argument types
	*/
	public Class[] getFormalArgumentTypes()  {
		
		try {
		Class[] clz = new Class[fargs.length];
		for(int i=0;i<clz.length;++i) 
			clz[i]=loadClass(fargs[i]);
		return clz;
		}catch (ClassNotFoundException ex) {
			jasco.util.logging.Logger.getInstance().showError(ex);
		}
		return new Class[0];
	}
	
	/**
	*  Returns the fully quantified name of class the method is declared
	*/
	public String getClassName()  {
	
		return className;
	}
	
	/**
	*  Returns the modifiers intger, see java.lang.reflect.Modifier
	*/
	public int getModifiers()  {
	
		return modifiers;
	}
	
	/**
	*  Returns an array of all arguments
	*/
	public Object[] getArgumentsArray() 
	{
		return args;
	}
	
	/**
	*  Do not use, internal usage only
	*/
	public void setArgumentsArray(Object[] v) 
	{
		args=v;
	}
	
	/**
	*  Do not use, internal usage only
	*/
	public void setCallbackObject(JAsCoBeanCallback callback) 
	{
		this.callback=callback;
	}
	
	/**
	*  Returns the object the method is invoked upon
	*/
	public Object getCalledObject() 
	{
		return calledobject;
	}
	
	/**
	*  Do not use, internal usage only
	*/
	public void setCalledObject(Object o) 
	{
		calledobject=o;
		
	}

	
	/**
	*  Invokes the replaced method or subsequent around advices with original arguments, only availabe in around advice. 
	*  Is equivalent to invoking proceed(), proceed is however the preferred because it will execute considerably faster because of optimizations.
	* 
	* @throws ProceedFinishedException proceed was previously invoked and all advices and original method already executed.
	*/
	public final Object invokeJAsCoMethod() throws Exception
	{	
		
		if (currentHook<linkedHooks.length)   {
				int hookIndex = currentHook++;
				Object o =  ((IHook)linkedHooks[hookIndex]).around(this,this.getCalledObject(),this.getArgumentsArray());
				return o;
		}
		else 
			return invokeOriginalJAsCoMethod();
	}
	
	/**
	*  Invokes the original method with original arguments, only availabe in around advice.
	*  WARNING: This method bypasses the around chain, subsequent arounds are not executed! Use invokeJAsCoMethod if you just want
	*  to proceed.
	*/
	public Object invokeOriginalJAsCoMethod() throws Exception
	{	
		return callback.dispatch(methodid,getCalledObject(),getArgumentsArray());
	}
	
	/**
	*  Invokes the replaced method or subsequent around advices, only availabe in around advice
	*  @param obj The object to invoke the method upon, is typically calledobject keyword
	*  @param args The arguments to invoke the proceed chain with.
	* 
	* @throws ProceedFinishedException proceed was previously invoked and all advices and original method already executed.
	*/
	public final Object invoke(Object obj, Object[] args) throws Exception
	{	
		if (currentHook<linkedHooks.length)   {
			int hookIndex = currentHook++;
			Object o =  ((IHook)linkedHooks[hookIndex]).around(this,this.getCalledObject(),args);
			return o; 
		}
		else 
			return invokeOriginal(obj,args);
	}
	
	/**
	*  Invokes the original method, only availabe in around advice.
	*  WARNING: This method bypasses the around chain, subsequent arounds are not executed! Use invokeJAsCoMethod if you just want
	*  to proceed.
	*  @param obj The object to invoke the method upon, is typically calledobject keyword
	*  @param args The arguments to invoke the original method with.
	*/
	public Object invokeOriginal(Object obj, Object[] args) throws Exception
	{	
		try {
			return callback.dispatch(methodid,obj,args);
		} catch(ArrayIndexOutOfBoundsException ex) {
			
			if(args.length!=fargs.length) {
				Logger.getInstance().showError("Illegal number of arguments for invoking the original behavior: "+args.length+" received while "+fargs.length+" expected!");
				return null;
			}
			else throw ex;
		}
	}
	
	private Vector transformArrayToVector(Object[] args)
	{
		Vector temp = new Vector();
		for (int i = 0; i < args.length ; i++) 
		{
			temp.addElement(args[i]);
		}
		return temp;
	}
	
	/**
	*  Do not use, internal usage only
	*/
	public void clearLinkedHooks()  {
		
		currentHook=0;
		linkedHooks=EMPTY_LINKED_HOOKS;
	}
	
	/**
	*  Do not use, internal usage only
	*/
	public void clearProceedCheck()  {
		
		currentHook=0;
	}
	
	
	/**
	*  Do not use, internal usage only
	*/
	public void setLinkedHooks(IHook[] link)  {
	
		linkedHooks=link;
		currentHook=0;
	}
	
	/**
	*  Do not use, internal usage only
	*/
	public IHook[] getLinkedHooks()  {
	
		return linkedHooks;
	}
	
	private boolean doNotCache = false;


	
	/**
	*  Do not use, internal usage only
	*/
	boolean doNotCache()  {
		return doNotCache;
	}
	
	/**
	*  Do not use, internal usage only
	*/
	void setDoNotCache(boolean b)  {
		doNotCache=b;
	}
	
	/**
	* Allows to invoke the original method all over again with original arguments. This means that the list of applicable
	* aspects is rebuild. WARNING: could easily cause an infinite loop if this aspect is applicable again! 
	* Does not work with method overloading yet.
	*/ 
	public Object invokeAgain()  throws Exception {
		
		return invokeAgain(getCalledObject(),getArgumentsArray());
	}
	
	/**
	* Allows to invoke the original method all over again. This means that the list of applicable
	* aspects is rebuild and all aspects (including the current one) are triggered again if applicable. 
	* WARNING: could easily cause an infinite loop if this aspect is applicable again! 
	* Does not work with method overloading yet.
	* @param calledobject The object to invoke the method upon, is typically calledobject keyword
	* @param args The arguments to invoke the original method with. 
	*/ 
	public Object invokeAgain(Object calledobject, Object[] args) throws Exception  {
	
		try   {
			Class clz = Class.forName(getClassName());
			java.lang.reflect.Method themethod = null;
			
			themethod=clz.getMethod(getName(),getFormalArgumentTypes());
			
			if (themethod==null)  {
			
				java.lang.reflect.Method[] methods = clz.getDeclaredMethods();
			
				for(int i=0; i<methods.length;++i)
					if (methods[i].getName().equals(getName()))  {
						themethod=methods[i];
						break;
					}
			}
					
			if(themethod==null)
				throw new IllegalArgumentException("again invocation for"+this+" failed (method not found)");
			
			resetProceedChain();
			Object result=themethod.invoke(getCalledObject(),args);	
			return result;
			
		}catch(InvocationTargetException ex)  {
			
			Throwable source = ex.getCause();
			if(source instanceof Exception)
				throw (Exception) source;
			else throw ex;
		}
	}
	
	/**
	* Allows to invoke the original method again with original arguments. If this aspect is however triggered by
	* a method defined in another aspect, the triggering method for that aspect is used unless that aspect is triggered by a method 
	* defined in an aspect etc...
	*/ 
	public Object invokeNoAspectAgain()  {
		
		throw new IllegalArgumentException("Method not yet implemented: invokeNoAspectAgain");
		/*
		if (getCalledObject() instanceof Cutpoint)  {
			Cutpoint cp = (Cutpoint) getCalledObject();
			JascoMethod method = cp._Jasco_getCalledMethod();
			if(method!=null)
				return method.invokeNoAspectAgain();
			else return invokeAgain();
		}
		else return invokeAgain();*/
	}	
	
	CombinedHooks getCombinedHooks()  {
	
		return hooksInvoker;
	}
	
	void setCombinedHooks(CombinedHooks c)  {
	
		hooksInvoker=c;
	}
	
	/**
	*  Do not use, internal usage only
	*/
	public boolean inlineCompiled()  {
	
		return inlineCompiled;
	}
	
	/**
	*  Do not use, internal usage only
	*/
	public void setInlineCompiled(boolean b)  {
		inlineCompiled=b;
	}

	/* (non-Javadoc)
	 * @see jasco.runtime.MethodJoinpoint#getExceptionTypes()
	 */
	public Class[] getExceptionTypes() {	
			try {
			Class[] clz = new Class[excs.length];
			for(int i=0;i<clz.length;++i) 
				clz[i]=loadClass(excs[i]);
			return clz;
			}catch (ClassNotFoundException ex) {
				jasco.util.logging.Logger.getInstance().showError(ex);
			}
			return new Class[0];
	}
	
	/**
	* Allows to reset the proceed chain, so that all applicable aspects (including the current one) are executed again when proceed is invoked.
	* In contary to invokeAgain, the list of appliable aspects is not rebuild.
	* WARNING: could easily cause an infinite loop if this aspect always invokes this method! 
	*/
	public final void resetProceedChain() throws Exception
	{	
		currentHook=0;
	}
	
	protected static final void initFrom(JascoMethod theMethod, CtMethod method,ClassLoader loader, int id) throws NotFoundException {
		
		theMethod.setReturnTypeName(method.getReturnType().getName());
		theMethod.dontLoadClass=true;
		
		CtClass[] exs = method.getExceptionTypes();
		String[] sexs = new String[exs.length];
		for(int i=0;i<exs.length;++i)
			sexs[i]=exs[i].getName();
		theMethod.setExceptionTypes(sexs);
		
		theMethod.initClassLoader(loader);
		
		CtClass[] fargs = method.getParameterTypes();
		String[] sfargs = new String[fargs.length];
		for(int i=0;i<fargs.length;++i)
			sfargs[i]=fargs[i].getName();
		theMethod.setFormalArgumentTypes(sfargs);
		
		theMethod.setReturnTypeName(method.getReturnType().getName());
		theMethod.line=method.getMethodInfo().getLineNumber(0);
		theMethod.javassistMethod=method;
	}
		
	public static JascoMethod buildFrom(CtMethod method,ClassLoader loader, int id) throws NotFoundException {
		
		String fullName = TransformerConstants.getFullName(method);
		
		JascoMethod theMethod = new JascoMethod(
				fullName,  
				method.getName(), 
				method.getDeclaringClass().getName(), 
				id, 
				null, 
				method.getModifiers(), 
				true);
		
		initFrom(theMethod,method,loader,id);
		
		return theMethod;
	}

	/* (non-Javadoc)
	 * @see jasco.runtime.MethodJoinpoint#getReturnTypeName()
	 */
	public String getReturnTypeName() {
		
		if(returnTypeName==null)
			try {
				returnTypeName=this.getJavassistMethod().getReturnType().getName();
			} catch (NotFoundException e) {
				Logger.getInstance().showError(e);
			}
		return returnTypeName; 
	}
	
	protected void setReturnTypeName(String s) {
		
		this.returnTypeName=s;
	}

	/* (non-Javadoc)
	 * @see jasco.runtime.MethodJoinpoint#getFormalArgumentTypeNames()
	 */
	public String[] getFormalArgumentTypeNames() {
		
		return fargs;
	}

	/**
	 * @return
	 */
	public boolean dontLoadClass() {
		return dontLoadClass;
	}
	
	public void dontLoadClass(boolean b) {
		dontLoadClass=b;
	}
	
	public JascoMethod copy() {
		
		try {
			return (JascoMethod) this.clone();
		} catch (CloneNotSupportedException e) {
			Logger.getInstance().showError(e);
		}
		return this;
	}
	
	public CtMethod getJavassistMethod() {
		
		if(this.javassistMethod==null)
			initJavassistMethod();
		return this.javassistMethod;
	}
	
	protected void initJavassistMethod() {
		
		try {
		String[] fargs = this.getFormalArgumentTypeNames();
		CtClass[] jfargs = new CtClass[fargs.length];
		for(int i=0;i<fargs.length;++i)
			jfargs[i]=JAsCoClassLoader.getSingleton().loadClass(fargs[i]);
		
		CtMethod method = JAsCoClassLoader.getSingleton().loadClass(getClassName()).getDeclaredMethod(getName(),jfargs);
		this.javassistMethod=method;
		}catch(Exception ex) {
			Logger.getInstance().showError(ex);
		}
		
		
	}

	public int getSourceLocation() {
		if(line<0)
			line= getJavassistMethod().getMethodInfo().getLineNumber(0);
		return line;
		
	}
}