/**
 * 
 */
package jasco.runtime.distribution;


//JGroups imports
import java.io.Serializable;

import jasco.runtime.ConnectorRegistry;
import jasco.runtime.JascoMethod;
import jasco.util.logging.Logger;

import org.jgroups.*;
import org.jgroups.blocks.*;
import org.jgroups.stack.IpAddress;
import org.jgroups.util.*;

/**
 * This singleton class manages the comunication of the Jasco
 * joinpoint(methodsInvocations) events with the group of 
 * instances that belongs 
 * to the same group.
 * 
 * @author lbenavid
 */
public class DistributePointcutHandler implements RequestHandler {
	private static DistributePointcutHandler _managerInstance;
	
	Channel            channel;
    MessageDispatcher  disp;
    RspList            rsp_list;
    String             props =   
    	DistributedOptions.getInstance().getStack("AspectsGroup");


	/**
	 * This class is a singleton class
	 */
	private DistributePointcutHandler() {
		try {
			start();
		} catch (Exception e) {
			Logger.getInstance().showDebug(e);
		}
	}
	
	public Channel getChannel() {
		
		return channel;
	}
	
    public void start() throws Exception {
        channel=new JChannel(props);
        channel.setOpt(Channel.LOCAL, false);//Not self delivery
        disp=new MessageDispatcher(channel, null, null, this);
        channel.connect("MessageDispatcherTestGroup");
        Logger.getInstance().showDebug(channel.getView().toString());
    }
	
	public static DistributePointcutHandler getInstance(){
		return _managerInstance;
	}
	
	/**Broad cast a message and do not care bout the answers*/
	public void broadcastMessageAsync(DistributedStandardMessage jm){
		RspList rsp_list=disp.castMessage(null,
                new Message(null, null, jm),
                GroupRequest.GET_NONE, 0);
	}
	
	/**Broad cast a message and waits for all the answers*/
	public RspList broadcastMessageSync(DistributedStandardMessage jm){
		RspList rsp_list=disp.castMessage(null,
                new Message(null, null, jm),
                GroupRequest.GET_ALL, 0);
		return rsp_list;
	}
	
	/**Broad cast a message and waits for the first answer*/
	public RspList broadcastMessageFirstAnswer(DistributedStandardMessage jm){
		RspList rsp_list=disp.castMessage(null,
                new Message(null, null, jm),
                GroupRequest.GET_FIRST, 0);
		return rsp_list;
	}
	
	/**Send a synchronous Message to an specific host*/
	public Object sendSyncMessage(DistributedStandardMessage jm, Host host){
		RspList rsp_list=disp.castMessage(null,
                new Message(host.getAddress(), null, jm),
                GroupRequest.GET_FIRST, 0 );
		if(rsp_list.size()<=0)
			throw new HostNotFoundException(host);
		else return ((Rsp) rsp_list.elementAt(0)).getValue(); 
	}
	
	public void sendNotification(DistributedJascoMethod jm){
		if(!jm.isRemoteJP()){
		   if(jm.getHost() == null)
			  jm.setHost(new Host((IpAddress)channel.getLocalAddress()));
		  jm.setRemoteJP(true);
		  Object tmpObject = jm.getCalledObject();
		  Object[] args = jm.getArgumentsArray();
		  Object[] argscopy = new Object[args.length];
		  for(int i=0;i<args.length;++i) {
			  argscopy[i]=args[i];
			  if(!(args[i] instanceof Serializable))
				  args[i]=null;
		  }
		  jm.setArgumentsArray(args);
		  if(!(tmpObject instanceof Serializable))
			  jm.setCalledObject(null);
	      broadcastMessageAsync(new DistributedStandardMessage(jm, DistributedStandardMessage.REMOTE_JP));
	      jm.setCalledObject(tmpObject);
	      jm.setArgumentsArray(argscopy);
	      jm.setRemoteJP(false);
		}
	}
	
	public Object invokeRemoteSynchronous(SynchronousAdviceExecution adviceExecution) {
		Host host = adviceExecution.getHost();
		return invokeRemoteSynchronousOn(host,adviceExecution);
	}
	
	public Object invokeRemoteSynchronousOn(Host host, SynchronousAdviceExecution adviceExecution) {
		
		//Modifies the Jp information to only send serializable information
		//Parameters are send by copy only when they are serializable
		//This solution is using the same object, this can be problematic for
		//concurrency, so in the future a clone of the jp can be used ldbn 22-9-2004
		DistributedJascoMethod jm = adviceExecution.getJP();
		Object tmpObject = jm.getCalledObject();
		Object[] args = jm.getArgumentsArray();
		Object[] argscopy = new Object[args.length];
		for(int i=0;i<args.length;++i) {
			argscopy[i]=args[i];
			if(!(args[i] instanceof Serializable))
				args[i]=null;
		}
		jm.setArgumentsArray(args);
		if(!(tmpObject instanceof Serializable))
		  jm.setCalledObject(null);
		
		//Make the remote call
		Object resp = sendSyncMessage(new DistributedStandardMessage(adviceExecution
				, DistributedStandardMessage.REMOTE_SYNC_MESSAGE), host);
		
		//restore jp values
	    jm.setCalledObject(tmpObject);
	    jm.setArgumentsArray(argscopy);

		return resp;
	}
	
	/**
	 * This methods execute the trap to the remote pointcut. 
	 * If the trap sends an error a false is answered, if the
	 * traps executes ok a true is answered. This can be changed
	 * to send objects, but for now there is no need for that.
	 * */
	public Object receiveNotification(JascoMethod jm){
		Object result;
		try {
			ConnectorRegistry.executeTrap(jm.getName(), jm.getCalledObject(), jm);
			result = true;
		} catch (Exception e) {
			Logger.getInstance().showDebug(e);
			result = false;
		}
		return result;
	}
	
	public Object handle(Message msg) {
		Logger.getInstance().showDebug(msg.toString());
		DistributedStandardMessage dsmsg = (DistributedStandardMessage) msg.getObject();
		if(dsmsg.getType() == DistributedStandardMessage.REMOTE_JP)
			return receiveNotification((JascoMethod) dsmsg.getMsg());
		if(dsmsg.getType() == DistributedStandardMessage.REMOTE_SYNC_MESSAGE)
			return DistributedSynchronizedInvoker.getInstance().invokeSynchronous(
					(SynchronousAdviceExecution) dsmsg.getMsg());
		if(dsmsg.getType() == DistributedStandardMessage.REMOTE_ASYNC_MESSAGE)
			return null;
		return null;
    }
	/**
	 * initializes this manager instance
	 *
	 */
	public static void init() {
		if(!initialized())
			_managerInstance=new DistributePointcutHandler();
		else throw new IllegalArgumentException(
				"Cannot initialize distribution manager more than once");
	}

	public static boolean initialized() {
		return _managerInstance!=null;
	}

}
