package jasco.runtime.distribution;

import jasco.options.Options;
import jasco.runtime.CRChangeEvent;
import jasco.runtime.ConnectorRegistry;
import jasco.runtime.ConnectorRegistryListener;
import jasco.runtime.aspect.IHook;
import jasco.runtime.connector.Connector;
import jasco.runtime.connector.HookElement;
import jasco.runtime.distribution.annotation.Distribution;
import jasco.runtime.distribution.annotation.JAsCoGenerated;
import jasco.runtime.distribution.annotation.Shared;
import jasco.util.generators.JavaGenerator;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.net.MalformedURLException;
import java.rmi.AlreadyBoundException;
import java.rmi.Naming;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.util.Enumeration;

public class DistributedSharingManager implements ConnectorRegistryListener {
	
	private static DistributedSharingManager _instance = new DistributedSharingManager();
	
	private DistributedSharingManager() {
		super();
		
		ConnectorRegistry.addConnectorRegistryListener(this);
	}
	
	public static DistributedSharingManager getInstance() {
		
		return _instance;
	}
	
	//TODO: remove sharing aspect when necessary (i.e. connector removed)...
	public void registryChanged(CRChangeEvent event) {
		
		if(event.getType()==CRChangeEvent.CONNECTOR_ADDED) {
			Connector c = (Connector) event.getParam();
			if(c.getClass().getAnnotation(JAsCoGenerated.class)==null)
				tackleSharing(c);
		}
		
	}

	private void tackleSharing(Connector c) {
		
		StringBuffer connectorCode = new StringBuffer();
		StringBuffer localConnectorCode = new StringBuffer();
		String localConnectorName = c.getName()+"_JAsCoDistributedSharingLocal";
		String globalConnectorName = c.getName()+"_JAsCoDistributedSharing";
		
		Enumeration enu = c.getHooks();
		while(enu.hasMoreElements()) {
			HookElement el = (HookElement) enu.nextElement();
			IHook hook = el.getCutpoint();
			tackleSharing(hook, hook.getClass(),connectorCode,localConnectorCode);
		}
		
		loadConnector(localConnectorName,localConnectorCode.toString(),true);
		loadConnector(globalConnectorName,connectorCode.toString(),false);
	}

	private void loadConnector(String connectorname, String impl, boolean isLocal) {
		
		if(impl.equals(""))
			return;
		
		String localString = "";
		if(isLocal)
			localString="@"+Distribution.class.getName()+"(value="+Distribution.class.getName()+".Type.LOCAL) ";

		impl=localString+"@"+JAsCoGenerated.class.getName()+" static connector "+connectorname+" {"+JavaGenerator.NEWLINE+impl+"}";
		
		File file = new File(connectorname+".con");
		try {
		
		
		FileWriter writer = new FileWriter(file);
		writer.write(impl);
		writer.close();
		
		} catch (IOException e) {
			jasco.util.logging.Logger.getInstance().showError("Connector failed writing: "+connectorname);
			jasco.util.logging.Logger.getInstance().showError(e);
		}
		
		String dir = Options.getOutputDir();
		try {
			Options.setOutputDir(Options.getTempDir().getAbsolutePath());
			jasco.Jasco.compileAndLoadConnector(file.getAbsolutePath());
		} catch (Throwable e) {
			jasco.util.logging.Logger.getInstance().showError("Connector failed loading: "+connectorname);
			jasco.util.logging.Logger.getInstance().showError(e);
		}
		finally{
			Options.setOutputDir(dir);
		}
		
		file.deleteOnExit();
		File classFileDir = new File(Options.getTempDir(),"Connector");
		File classFile = new File(classFileDir,connectorname+".class");
		if(classFile.exists())
			classFile.deleteOnExit();
		
	}
	
	private void tackleSharing(IHook hook, Class<? extends IHook> hookClass, StringBuffer connectorCode,StringBuffer localConnectorCode) {
		Field[] fields = hookClass.getFields();
		for(int i=0;i<fields.length;++i)
			tackleSharing(hook,fields[i],connectorCode,localConnectorCode);
	}

	private void tackleSharing(IHook hook, Field field, StringBuffer connectorCode,StringBuffer localConnectorCode) {
		Annotation annon = field.getAnnotation(Shared.class);
		Shared sharingMod = (Shared) annon;
		if(sharingMod == null)
			return;
		
		boolean masterExists = checkMasterExists(field);
		
		if(!masterExists)
			deployMaster(hook,field);
		if(sharingMod.type()==Shared.Type.GLOBAL&&!masterExists)
			generateGlobalConnectorCode(field,connectorCode);
		
		generateLocalReference(field,localConnectorCode);
	}
	
	private void deployMaster(IHook hook, Field field) {
		
		try {
		field.setAccessible(true);
		}catch(SecurityException ex) {
			jasco.util.logging.Logger.getInstance().showError("Cannot access field "+field.getName()+" in aspect "+hook.getClass().getName());
		}
		Object value = null;
		try {
			value = field.get(hook);
		} catch (IllegalArgumentException e1) {
			jasco.util.logging.Logger.getInstance().showError(e1);
		} catch (IllegalAccessException e1) {
			jasco.util.logging.Logger.getInstance().showError("Cannot access field "+field.getName()+" in aspect "+hook.getClass().getName());
		}
		
		
		try {
			DefaultSharedField sfield = new DefaultSharedField(value);
			Naming.bind(getMasterName(field),sfield);
		} catch (MalformedURLException e) {
			jasco.util.logging.Logger.getInstance().showError(e);
		} catch (RemoteException e) {
			jasco.util.logging.Logger.getInstance().showError("Cannot connect to registry, is rmiregistry running?");
			jasco.util.logging.Logger.getInstance().showDebug(e);
		} catch (AlreadyBoundException e) {
			jasco.util.logging.Logger.getInstance().showError(e);
		}
	}
	
	private boolean checkMasterExists(Field field) {
		
		return checkMasterExists(getMasterName(field));
	}

	private boolean checkMasterExists(String masterName) {
		
		try {
			Object bla = Naming.lookup("//localhost/"+masterName);
			return true;
		} catch (MalformedURLException e) {
			jasco.util.logging.Logger.getInstance().showError(e);
		} catch (RemoteException e) {
			jasco.util.logging.Logger.getInstance().showError(e);
		} catch (NotBoundException e) {
			return false;
		}
		
		return false;
	}
	
	private String getMasterName(Field field) {
		
		return getMasterName(field.getDeclaringClass().getName(),field.getName());
	}
	
	private String getMasterName(String aspectname, String fieldname) {
		
		return "jasco.distribution.sharing."+aspectname+"."+fieldname;
	}
//	 TODO Auto-generated catch block
	private void generateLocalReference(Field field, StringBuffer connectorCode) {
		
		String aspectname = "jasco.runtime.distribution.aspects.Sharing.QueryMaster";
		generateConnectorCode(aspectname,getGetterName(field),field,connectorCode);
		
		aspectname = "jasco.runtime.distribution.aspects.Sharing.SetMaster";
		generateConnectorCode(aspectname,getSetterName(field),field,connectorCode);
		
	}
	
	private void generateConnectorCode(String aspectname,String fieldname,Field field,StringBuffer res) {
		
			
			String var = newUniqueVar();
			
			res.append(aspectname+" "+var+" = new "+aspectname+"(* ");
			res.append(field.getDeclaringClass().getName()+"."+fieldname+"(*)");
			res.append(");"+JavaGenerator.NEWLINE);
			
			res.append(var+".setSharedFieldInfo(");
			res.append("\""+field.getDeclaringClass().getName()+"\", \""+field.getName()+"\");"+JavaGenerator.NEWLINE);
			
	}

	private void generateGlobalConnectorCode(Field field, StringBuffer connectorCode) {
		
		String aspectname = "jasco.runtime.distribution.aspects.Sharing.Synchronize";
		String var = newUniqueVar();
		
		connectorCode.append(aspectname+" "+var+" = new "+aspectname+"(* ");
		connectorCode.append(field.getDeclaringClass().getName()+"."+getSetterName(field)+"(*)");
		connectorCode.append(");"+JavaGenerator.NEWLINE);

	}
	
	private static int id=0;
	
	private String newUniqueVar() {
		return "hook"+id++;
	}

	public String getSetterName(Field field) {
		
		return "set"+JavaGenerator.upperCaseFirst(field.getName());
	}
	
	public String getGetterName(Field field) {
		
		return "get"+JavaGenerator.upperCaseFirst(field.getName());
	}

	public void registryUpdated() {
		
		
	}
	
	public ISharedField getSharedMasterField(String aspectname, String fieldname) {
		
		String mastername = getMasterName(aspectname,fieldname);

		try {
			ISharedField bla = (ISharedField) Naming.lookup("//localhost/"+mastername);
			return bla;
		} catch (MalformedURLException e) {
			jasco.util.logging.Logger.getInstance().showError(e);
			return  null;
		} catch (RemoteException e) {
			jasco.util.logging.Logger.getInstance().showError(e);
			return  null;
		} catch (NotBoundException e) {
			jasco.util.logging.Logger.getInstance().showError(e);
		}
		
		
		return null;
	}
}
